/*
 *  (Gnono) Card game for GNOME
 *
 *  Copyright (C) 2001-2002 Dennis J Houy <djhouy@paw.co.za>
 *            (C) 2006      Yaakov Selkowitz <yselkowitz@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *  $Id: game.c,v 1.5 2007/01/04 06:09:00 yselkowitz Exp $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/garray.h>
#include <glib/gmessages.h>
#include <glib/gtypes.h>
#include <gtk/gtkmain.h>
#include <libgnome/gnome-config.h>

#include "ai.h"
#include "cards.h"
#include "interface.h"
#include "players.h"
#include "game.h"

static const gint NUM_PLAYERS = 2;

gint _cards_to_deal;
gint _playout_score;

gboolean _game_in_progress = FALSE;
gboolean _card_played = FALSE;
gboolean _force_end_game = FALSE;
gboolean _discard_applied = FALSE;

void apply_discard (void);

void
deal_cards ()
{
    gint c, p;

    for (c = 0; c < _cards_to_deal; c++) {
        for (p = 0; p < NUM_PLAYERS; p++) {
            player_cards_add (p, deck_take_top ());
        }
    }
}

gboolean
get_cards_to_deal ()
{
    return _cards_to_deal;
}

gint
get_num_players ()
{
    return NUM_PLAYERS;
}

gboolean
get_playout_score (void)
{
    return _playout_score;
}

void
play_card (gint player, gint card_idx)
{
    Card *play_card;

    play_card = player_cards_remove_index (player, card_idx);

    discard_add_top (play_card);

    _discard_applied = FALSE;

    if (get_player (get_current_player ())->cards->len == 0)
        _game_in_progress = FALSE;

    if ((get_player (player)->type == GN_PLAYER_TYPE_HUMAN) && _game_in_progress &&
        ((play_card->type == GN_CARD_WILD) || (play_card->type == GN_CARD_WILD_DRAW_4))) {
        gci_color_picker_show ();
    } else {
        _card_played = TRUE;
    }
}

void
set_card_played (gboolean played)
{
    _card_played = played;
}

gboolean
valid_play (gint player, gint card_idx)
{
    Card *test_card, *play_card;
    gint c;

    play_card = (Card*) g_ptr_array_index (get_player (player)->cards, card_idx);

    if (play_card->type == GN_CARD_WILD) {
        return TRUE;
    } else if (play_card->type == GN_CARD_WILD_DRAW_4) {
        for (c = 0; c < get_player (player)->cards->len; c++) {
            test_card = (Card*) g_ptr_array_index (get_player (player)->cards, c);
            if (test_card->color == discard_top_card ()->color) return FALSE;
        }
        return TRUE;
    } else if ((play_card->type == discard_top_card ()->type) ||
               (play_card->color == discard_top_card ()->color)) {
        return TRUE;
    }

    return FALSE;
}

void
game_begin ()
{
    gint p;

#ifdef DEBUG
    g_message ("%s", "Entered game_begin ()");
#endif

    _cards_to_deal = gnome_config_private_get_int ("Gnono/GamePlay/CardsToDeal=7");
    _playout_score = gnome_config_private_get_int ("Gnono/GamePlay/PlayoutScore=100");

    panels_init ();
    players_init ();

    _force_end_game = FALSE;

    set_current_player (0);

    while (!_force_end_game) {
        for (p = 0; p < get_num_players (); p++) {
            player_cards_remove_all (p);
        }

        deck_init ();
        discard_init ();
        deal_cards ();

        _game_in_progress = TRUE;
        _discard_applied = FALSE;

        apply_discard ();
    
        display_starting_player (get_current_player ());

#ifdef DEBUG
        g_message ("%s", "Entering game loop now");
#endif

        do {
            gtk_main_iteration_do (FALSE);

            _card_played = FALSE;

            do {
                if (get_player (get_current_player ())->type == GN_PLAYER_TYPE_COMPUTER) {
                    computer_play (TRUE);
                    _card_played = TRUE;
                } else {
                    gtk_main_iteration_do (FALSE);
                }
            } while (!_card_played && !_force_end_game);

            apply_discard ();
        } while (_game_in_progress && !_force_end_game);

#ifdef DEBUG
        g_message ("%s", "Just returned from game loop");
#endif

        if (_force_end_game) _game_in_progress = FALSE;

        if (!_force_end_game) {
            for (p = 0; p < get_num_players (); p++) {
                if (p != get_current_player ()) {
                    get_player (get_current_player ())->score += get_hand_score (p);
                }
            }

            update_score_card (get_current_player ());

            gci_discard_show_color ();

            if (get_player (get_current_player ())->score >= _playout_score) {
                display_winning_player (get_current_player ());
                _force_end_game = TRUE;
            } else {
                prompt_to_continue_game (get_current_player ());
            }
        }
    }

    if (app_is_closing ())
        app_destroy ();

#ifdef DEBUG
    g_message ("%s", "Exiting game_begin ()");
#endif
}

void
apply_discard (void)
{
    gint c, p;
    Card *played_card;

    if (_force_end_game) return;

    gci_discard_show_color ();

    played_card = discard_top_card ();

    if (_discard_applied) {
        if (_game_in_progress)
            set_current_player (next_player (get_current_player ()));
    } else {
        if (played_card->type == GN_CARD_WILD_DRAW_4) {
            p = next_player (get_current_player ());
            for (c = 0; c < 4; c++) {
                player_cards_add (p, deck_take_top ());
            }
        } else if (played_card->type == GN_CARD_TAKE_2) {
            p = next_player (get_current_player ());
            for (c = 0; c < 2; c++) {
                player_cards_add (p, deck_take_top ());
            }
        } else if (played_card->type == GN_CARD_REVERSE) {
            if (get_num_players () > 2) {
                players_reverse_direction ();
                if (_game_in_progress)
                    set_current_player (next_player (get_current_player ()));

            } else { /* behaves the same as a Skip card */
                p = next_player (get_current_player ());
                if (_game_in_progress)
                    set_current_player (next_player (p));
            }

        } else if (played_card->type == GN_CARD_SKIP) {
            p = next_player (get_current_player ());
            if (_game_in_progress)
                set_current_player (next_player (p));
        } else {
            if (_game_in_progress)
                set_current_player (next_player (get_current_player ()));
        }
    }

    _discard_applied = TRUE;
}

gboolean
game_in_progress ()
{
    return _game_in_progress;
}

void
force_end_game (void)
{
    _force_end_game = TRUE;

#ifdef DEBUG
    g_message ("%s", "Issued game end command");
#endif
}
