/*
 *  (Gnono) Card game for GNOME
 *
 *  Copyright (C) 2001-2002 Dennis J Houy <djhouy@paw.co.za>
 *            (C) 2006      Yaakov Selkowitz <yselkowitz@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *  $Id: interface.c,v 1.12 2007/01/04 06:21:37 yselkowitz Exp $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <time.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gdk-pixbuf/gdk-pixbuf-core.h>
#include <gdk/gdkevents.h>
#include <gdk/gdkrgb.h>
#include <gtk/gtk.h>
#include <libgnomecanvas/gnome-canvas.h>
#include <libgnomecanvas/gnome-canvas-pixbuf.h>
#include <libgnomecanvas/gnome-canvas-rect-ellipse.h>
#include <bonobo/bonobo-dock.h>
#include <libgnomeui/gnome-app-helper.h>
#include <libgnomeui/gnome-messagebox.h>
#include <libgnomeui/gnome-ui-init.h>

#include "cards.h"
#include "ai.h"
#include "game.h"
#include "interface.h"
#include "players.h"
#include "preferences.h"

gboolean _app_is_closing = FALSE;

static GtkWidget *gn_app = NULL;
static GnomeCanvas *gn_canvas = NULL;
static GnomeCanvasItem *gci_pile_pick_up = NULL;
static GnomeCanvasItem *gci_pile_discard = NULL;
static GnomeCanvasItem *gci_discard_color = NULL;
static GnomeCanvasGroup *gn_pnl_color_picker = NULL;
static GtkWidget *gn_score_column = NULL;

void panels_init (void);
void score_column_destroy (void);
void toolbar_set_style (void);
void about_cb (GtkWidget *widget, gpointer data);
gint game_quit_cb (GtkWidget *widget, gpointer data);
gint gci_hand_cards_clicked_cb (GnomeCanvasItem *i, GdkEvent *event, gpointer data);
gint gci_pnl_color_picker_cb (GnomeCanvasItem *item, GdkEvent *event, gpointer data);
gint gci_pile_pick_up_clicked_cb (GnomeCanvasItem *item, GdkEvent *event, gpointer data);
gint make_card_click_data (gint player, gint card);

static GnomeUIInfo game_menu[] = {
    GNOMEUIINFO_MENU_NEW_GAME_ITEM (game_new_cb, NULL),
    GNOMEUIINFO_SEPARATOR,
    GNOMEUIINFO_MENU_PREFERENCES_ITEM (dlg_preferences, NULL),
    GNOMEUIINFO_SEPARATOR,
    GNOMEUIINFO_MENU_EXIT_ITEM (game_quit_cb, NULL),
    GNOMEUIINFO_END
};

static GnomeUIInfo help_menu[] = {
    GNOMEUIINFO_MENU_ABOUT_ITEM (about_cb, NULL),
    GNOMEUIINFO_END
};

static GnomeUIInfo main_menu[] = {
    GNOMEUIINFO_MENU_GAME_TREE (game_menu),
    GNOMEUIINFO_MENU_HELP_TREE (help_menu),
    GNOMEUIINFO_END
};

static GnomeUIInfo main_toolbar[] = {
    GNOMEUIINFO_ITEM_STOCK (N_("New Game"), N_("Start a new game"),
                            game_new_cb, GTK_STOCK_NEW),
    GNOMEUIINFO_ITEM_STOCK (N_("Preferences"), N_("Edit preferences"),
                            dlg_preferences, GTK_STOCK_PREFERENCES),
    GNOMEUIINFO_ITEM_STOCK (N_("Exit"), N_("Exit Gnono"),
                            game_quit_cb, GTK_STOCK_QUIT),
    GNOMEUIINFO_END
};

GtkWindow *
main_app_window (void)
{
    return GTK_WINDOW (gn_app);
}

gboolean
app_is_closing (void)
{
    return _app_is_closing;
}

void
app_init (int argc, char *argv[])
{
    GtkWidget *gn_appbar;
    GtkWidget *gn_hbox;
    GtkWidget *gn_vseparator;
    GtkWidget *gn_hseparator;

    gnome_init ("Gnono", VERSION, argc, argv);

    gn_app = gnome_app_new ("gnono", _("Gnono"));
	gtk_window_set_default_icon_name ("gnono");

    gtk_window_set_resizable (GTK_WINDOW (gn_app), FALSE);

    g_signal_connect (GTK_OBJECT (gn_app), "delete_event",
                        GTK_SIGNAL_FUNC (game_quit_cb),
                        NULL);

    /* Create the menu bar */
    gnome_app_create_menus (GNOME_APP (gn_app), main_menu);

    /* Create the toolbar */
    gnome_app_create_toolbar (GNOME_APP (gn_app), main_toolbar);
    toolbar_set_style ();

    /* Create the status bar */
    gn_appbar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_USER);
    gnome_app_set_statusbar (GNOME_APP (gn_app), gn_appbar);
    gnome_app_install_appbar_menu_hints (GNOME_APPBAR (gn_appbar),  main_menu);

    /* Create the canvas */
    // gtk_widget_push_visual (gdk_rgb_get_visual ());
    gtk_widget_push_colormap (gdk_rgb_get_cmap ());
    gn_canvas = GNOME_CANVAS (gnome_canvas_new ());
    gtk_widget_pop_colormap ();
    //gtk_widget_pop_visual ();

    gnome_canvas_set_scroll_region (gn_canvas, 0.0, 0.0, 500.0, 350.0);

    gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                           gnome_canvas_rect_get_type (),
                           "x1", 0.0,
                           "y1", 0.0,
                           "x2", 500.0,
                           "y2", 350.0,
                           "fill_color", "darkgreen",
                           NULL);

    gtk_widget_set_size_request (GTK_WIDGET (gn_canvas), 500, 350);

    gn_hbox = gtk_hbox_new (FALSE, 0);

    gtk_box_pack_start (GTK_BOX (gn_hbox),
                        GTK_WIDGET (gn_canvas),
                        FALSE, FALSE, 0);

    gn_vseparator = gtk_vseparator_new ();

    gtk_box_pack_start (GTK_BOX (gn_hbox),
                        GTK_WIDGET (gn_vseparator),
                        TRUE, TRUE, 1);

    gn_score_column = gtk_vbox_new (FALSE, 0);

    gtk_box_pack_start (GTK_BOX (gn_hbox),
                        GTK_WIDGET (gn_score_column),
                        FALSE, FALSE, 0);

    gn_hseparator = gtk_hseparator_new ();

    gtk_widget_set_size_request (GTK_WIDGET (gn_hseparator), 124, 2);

    gtk_box_pack_start (GTK_BOX (gn_score_column),
                        GTK_WIDGET (gn_hseparator),
                        FALSE, TRUE, 0);

    gn_hseparator = gtk_hseparator_new ();

    gtk_widget_set_size_request (GTK_WIDGET (gn_hseparator), 124, 2);

    gtk_box_pack_start (GTK_BOX (gn_score_column),
                        GTK_WIDGET (gn_hseparator),
                        FALSE, TRUE, 0);

    gn_hseparator = gtk_hseparator_new ();

    gtk_widget_set_size_request (GTK_WIDGET (gn_hseparator), 124, 2);

    gtk_box_pack_start (GTK_BOX (gn_score_column),
                        GTK_WIDGET (gn_hseparator),
                        FALSE, TRUE, 0);

    gnome_app_set_contents (GNOME_APP (gn_app), GTK_WIDGET (gn_hbox));

    gtk_widget_show_all (gn_app);

    srand (time (0));

    panels_init ();
}

void
app_destroy (void)
{
#ifdef DEBUG
    g_message ("%s", "Entered app_destroy ()");
#endif

    if (game_in_progress ()) {
        _app_is_closing = TRUE;
        force_end_game ();

#ifdef DEBUG
        g_message ("%s", "Game in progress, exiting for cleanup");
#endif

        return;
    }

    players_destroy ();
    deck_destroy ();
    gtk_widget_destroy (gn_app);
    gtk_main_quit ();

#ifdef DEBUG
    g_message ("%s", "This is the end");
#endif
}

void
toolbar_set_style (void)
{
    BonoboDockItem *item;
    GtkWidget *toolbar;

    item = gnome_app_get_dock_item_by_name (GNOME_APP (gn_app),
                                            GNOME_APP_TOOLBAR_NAME);
    toolbar = bonobo_dock_item_get_child (item);
    gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_BOTH);
    gtk_widget_queue_resize (gn_app);
}

void
about_cb (GtkWidget *widget, gpointer data)
{
	GdkPixbuf *logo = gdk_pixbuf_new_from_file(PIXMAPDIR "/gnono-about.png", NULL);

    const gchar *authors[] = {
        "Yaakov Selkowitz <yselkowitz@users.sourceforge.net>",
		"Daud Zoss <daudzoss@users.sourceforge.net>",
        "Dennis J Houy <djhouy@paw.co.za>",
        "Erlend Nagel <erlend_nagel@yahoo.com>",
        NULL};

	const gchar license[] = \
      	N_("This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.\n\n"
         "This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.\n\n"
         "You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.");

    const gchar translators[] = \
        "Erlend Nagel <erlend_nagel@yahoo.com>\n";

    gtk_show_about_dialog (NULL, 
	                       "authors", authors,
	                       "comments", _("A new style of card game for Gnome."),
	                       "copyright", "\xc2\xa9 2001-2002 PAW Digital Dynamics\n\xc2\xa9 2006 Yaakov Selkowitz",
	                       "license", license,
	                       "logo", logo,
	                       "translator-credits", translators,
	                       "version", VERSION,
	                       "website", "http://sourceforge.net/projects/gnono/",
	                       "wrap-license", TRUE,
	                       NULL);
}

gint
game_new_cb (GtkWidget *widget, gpointer data)
{
    static gboolean _restart = TRUE;

    if (game_in_progress ()) {
        if (!_app_is_closing) {
            if (!prompt_user_question (_("Do you want to end the current game\nand start a new one?")))
                return TRUE;
        }

        force_end_game ();
        _restart = TRUE;
        return TRUE;
    }

    while (!_app_is_closing) {
        if (!_restart) break;

        _restart = FALSE;
        game_begin ();
    }

    _restart = TRUE;

    return TRUE;
}

gint
game_quit_cb (GtkWidget *widget, gpointer data)
{
    if (game_in_progress ()) {
        if (!prompt_user_question (_("Do you want to quit the current game and exit?")))
            return TRUE;
    }

    app_destroy ();

    return TRUE;
}

void
prompt_user_info (gchar *prompt, gchar *button)
{
    GtkWidget *dialog;

    dialog = gnome_message_box_new (prompt,
                                    GTK_STOCK_DIALOG_INFO,
                                    button,
                                    NULL);

    gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (gn_app));
    gnome_dialog_set_default (GNOME_DIALOG (dialog), 0);
    gnome_dialog_set_close (GNOME_DIALOG (dialog), TRUE);
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
}

gboolean
prompt_user_question (gchar *prompt)
{
    GtkWidget *dialog;

    dialog = gnome_message_box_new (prompt,
                                    GTK_STOCK_DIALOG_QUESTION,
                                    GTK_STOCK_YES,
                                    GTK_STOCK_NO,
                                    NULL);

    gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (gn_app));
    gnome_dialog_set_default (GNOME_DIALOG (dialog), 0);
    gnome_dialog_set_close (GNOME_DIALOG (dialog), TRUE);
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);

    return (gnome_dialog_run_and_close (GNOME_DIALOG (dialog)) == 0);
}

void
panels_init (void)
{
    GnomeCanvasItem *item;

    if (gn_pnl_color_picker != NULL) {
        gci_color_picker_hide ();
        return;
    }

    item = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                                  gnome_canvas_group_get_type (),
                                  "x", 220.0,
                                  "y", 20.0,
                                  NULL);

    gn_pnl_color_picker = GNOME_CANVAS_GROUP (item);

    gnome_canvas_item_new (gn_pnl_color_picker,
                           gnome_canvas_rect_get_type (),
                           "x1", 0.0,
                           "y1", 0.0,
                           "x2", 220.0,
                           "y2", 60.0,
                           "fill_color", "gray",
                           NULL);

    item = gnome_canvas_item_new (gn_pnl_color_picker,
                                  gnome_canvas_rect_get_type (),
                                  "x1", 20.0,
                                  "y1", 20.0,
                                  "x2", 50.0,
                                  "y2", 40.0,
                                  "fill_color", "red",
                                  NULL);

    g_signal_connect (GTK_OBJECT (item), "event",
                        GTK_SIGNAL_FUNC (gci_pnl_color_picker_cb),
                        GINT_TO_POINTER (GN_COLOR_RED));

    item = gnome_canvas_item_new (gn_pnl_color_picker,
                           gnome_canvas_rect_get_type (),
                           "x1", 70.0,
                           "y1", 20.0,
                           "x2", 100.0,
                           "y2", 40.0,
                           "fill_color", "darkgreen",
                           NULL);

    g_signal_connect (GTK_OBJECT (item), "event",
                        GTK_SIGNAL_FUNC (gci_pnl_color_picker_cb),
                        GINT_TO_POINTER (GN_COLOR_GREEN));

    item = gnome_canvas_item_new (gn_pnl_color_picker,
                           gnome_canvas_rect_get_type (),
                           "x1", 120.0,
                           "y1", 20.0,
                           "x2", 150.0,
                           "y2", 40.0,
                           "fill_color", "blue",
                           NULL);

    g_signal_connect (GTK_OBJECT (item), "event",
                        GTK_SIGNAL_FUNC (gci_pnl_color_picker_cb),
                        GINT_TO_POINTER (GN_COLOR_BLUE));

    item = gnome_canvas_item_new (gn_pnl_color_picker,
                           gnome_canvas_rect_get_type (),
                           "x1", 170.0,
                           "y1", 20.0,
                           "x2", 200.0,
                           "y2", 40.0,
                           "fill_color", "yellow",
                           NULL);

    g_signal_connect (GTK_OBJECT (item), "event",
                        GTK_SIGNAL_FUNC (gci_pnl_color_picker_cb),
                        GINT_TO_POINTER (GN_COLOR_YELLOW));

    gci_color_picker_hide ();
}

GtkWidget *
score_card_playout (void) {
    GtkWidget *playout_score;
    GtkWidget *item;
    GtkWidget *table;

    playout_score = g_new (GtkWidget, 1);

    table = gtk_table_new (1, 2, FALSE);

    item = gtk_label_new (_("Points to Win:"));
    gtk_misc_set_alignment (GTK_MISC (item), 0.0, 0.5);
    gtk_widget_show (item);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (item),
                      0, 1, 0, 1,
                      GTK_FILL,
                      0,
                      2, 2);

    playout_score = gtk_label_new (NULL);
    gtk_misc_set_alignment (GTK_MISC (playout_score), 1.0, 0.5);
    gtk_widget_show (playout_score);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (playout_score),
                      1, 2, 0, 1,
                      GTK_FILL,
                      0,
                      2, 2);

    item = gtk_hseparator_new ();
    gtk_widget_show (item);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (item),
                      0, 2, 1, 2,
                      GTK_EXPAND + GTK_FILL,
                      GTK_FILL,
                      2, 0);

    gtk_box_pack_start (GTK_BOX (gn_score_column),
                        GTK_WIDGET (table),
                        FALSE, FALSE, 0);

    gtk_widget_show (table);

	return (playout_score);
}

ScoreCard *
gn_score_card_new (gint player) {
    ScoreCard *new_score_card;
    GtkWidget *item;
    GtkWidget *table;
    gchar *underline;

    new_score_card = g_new (ScoreCard, 1);

    table = gtk_table_new (4, 2, FALSE);

    item = gtk_label_new (get_player (player)->name);

    underline = g_strnfill (strlen (get_player (player)->name), '_');
    gtk_label_set_pattern (GTK_LABEL (item), underline);
    g_free (underline);

    gtk_misc_set_alignment (GTK_MISC (item), 0.0, 0.5);
    gtk_widget_show (item);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (item),
                      0, 2, 0, 1,
                      GTK_FILL,
                      0,
                      2, 2);


    item = gtk_label_new (_("Total Points:"));
    gtk_misc_set_alignment (GTK_MISC (item), 0.0, 0.5);
    gtk_widget_show (item);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (item),
                      0, 1, 2, 3,
                      GTK_FILL,
                      0,
                      2, 2);

    new_score_card->total_points = gtk_label_new ("0");
    gtk_misc_set_alignment (GTK_MISC (new_score_card->total_points), 1.0, 0.5);
    gtk_widget_show (new_score_card->total_points);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (new_score_card->total_points),
                      1, 2, 2, 3,
                      GTK_FILL,
                      0,
                      2, 2);

    item = gtk_hseparator_new ();
    gtk_widget_show (item);

    gtk_table_attach (GTK_TABLE (table),
                      GTK_WIDGET (item),
                      0, 2, 3, 4,
                      GTK_EXPAND + GTK_FILL,
                      GTK_FILL,
                      2, 0);

    gtk_box_pack_start (GTK_BOX (gn_score_column),
                        GTK_WIDGET (table),
                        FALSE, FALSE, 0);

    gtk_widget_show (table);

    return (new_score_card);
}

gint
gci_pnl_color_picker_cb (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
    if (event->type == GDK_BUTTON_PRESS) {
        discard_top_card ()->color = GPOINTER_TO_INT (data);

        gnome_canvas_item_hide (GNOME_CANVAS_ITEM (gn_pnl_color_picker));

        set_card_played (TRUE);
    }

    return FALSE;
}

gint
gci_pile_pick_up_clicked_cb (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
    Player *p;

    if (!game_in_progress ()) return TRUE;

    if (event->type == GDK_BUTTON_PRESS) {
        player_cards_add (get_current_player (), deck_take_top ());

        p = get_player (get_current_player ());

        if (valid_play (get_current_player (), p->cards->len - 1)) {
            if (prompt_user_question (_("Do you want to play this card?"))) {
                play_card (get_current_player (), p->cards->len - 1);
                return TRUE;
            }
        }

        set_card_played (TRUE);
        return TRUE;
    }

    return FALSE;
}

void
display_winning_player (gint player)
{
    gchar *msg;

    msg = g_strdup_printf (_("%s wins this game. %s!"), get_player (player)->name,
            get_player (player)->type == GN_PLAYER_TYPE_HUMAN ? _("Well done") : _("Too bad"));

    prompt_user_info (msg, GTK_STOCK_OK);

    g_free (msg);
}

void
prompt_to_continue_game (gint player)
{
    gchar *msg;

    msg = g_strdup_printf (_("%s wins this round."), get_player (player)->name);

    prompt_user_info (msg, _("Continue"));

    g_free (msg);
}

void
display_starting_player (gint player)
{
    gchar *msg;

    msg = g_strdup_printf (_("%s starts."), get_player (player)->name);

    prompt_user_info (msg, GTK_STOCK_OK);

    g_free (msg);
}

void
gci_pick_up_set_image ()
{
    GdkPixbuf *p;

    if (gci_pile_pick_up == NULL ) {
        gci_pile_pick_up = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                                                  gnome_canvas_pixbuf_get_type (),
                                                  "x", 10.0,
                                                  "y", 10.0,
                                                  "width", GN_CARD_WIDTH,
                                                  "height", GN_CARD_HEIGHT,
                                                  NULL);

        g_signal_connect (GTK_OBJECT (gci_pile_pick_up), "event",
                            GTK_SIGNAL_FUNC (gci_pile_pick_up_clicked_cb), NULL);
    }

    if (deck_length () == 0) {
        p = card_get_picture (GN_COLOR_NONE, GN_CARD_EMPTY);
    } else {
        p = card_get_picture (GN_COLOR_NONE, GN_CARD_BACK);
    }

    gnome_canvas_item_set (gci_pile_pick_up, "pixbuf", p, NULL);

    gdk_pixbuf_unref (p);
}

void
gci_discard_set_image ()
{
    GdkPixbuf *p;

    if (gci_pile_discard == NULL) {
        gci_pile_discard = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                                                  gnome_canvas_pixbuf_get_type (),
                                                  "x", 88.0,
                                                  "y", 10.0,
                                                  "width", GN_CARD_WIDTH,
                                                  "height", GN_CARD_HEIGHT,
                                                  NULL);
    }

    p = card_get_picture (discard_top_card ()->color, discard_top_card ()->type);

    gnome_canvas_item_set (gci_pile_discard, "pixbuf", p, NULL);

    gdk_pixbuf_unref (p);
}

void
gci_discard_show_color ()
{
    if (gci_discard_color == NULL) {
        gci_discard_color = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                       gnome_canvas_rect_get_type (),
                       "x1", 85.0,
                       "y1", 7.0,
                       "x2", 159.0,
                       "y2", 111.0,
                       "fill_color", "black",
                       NULL);
        gnome_canvas_item_lower_to_bottom (gci_discard_color);
        gnome_canvas_item_raise (gci_discard_color, 1);
    }

    if (((discard_top_card ()->type == GN_CARD_WILD) ||
        (discard_top_card ()->type == GN_CARD_WILD_DRAW_4)) && game_in_progress ()) {

        switch (discard_top_card ()->color) {
        case GN_COLOR_NONE:
            gnome_canvas_item_set (gci_discard_color, "fill_color", "black", NULL);
            break;
        case GN_COLOR_RED:
            gnome_canvas_item_set (gci_discard_color, "fill_color", "red", NULL);
            break;
        case GN_COLOR_GREEN:
            gnome_canvas_item_set (gci_discard_color, "fill_color", "green", NULL);
            break;
        case GN_COLOR_BLUE:
            gnome_canvas_item_set (gci_discard_color, "fill_color", "blue", NULL);
            break;
        case GN_COLOR_YELLOW:
            gnome_canvas_item_set (gci_discard_color, "fill_color", "yellow", NULL);
            break;
        }

        gnome_canvas_item_show (gci_discard_color);
    } else {
        gnome_canvas_item_hide (gci_discard_color);
    }
}

void
gci_hand_cards_init (gint player)
{
    GnomeCanvasItem *gci_new;
    GdkPixbuf *p;

    if (get_player (player)->gci_hand_cards != NULL)
        gci_hand_cards_destroy (player);

    get_player (player)->gci_hand_cards = g_ptr_array_new ();

    p = card_get_picture (GN_COLOR_NONE, GN_CARD_EMPTY);

    gci_new = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                                     gnome_canvas_pixbuf_get_type (),
                                     "pixbuf", p,
                                     "x", 10.0,
                                     "y", 30.0 + (105.0 * (player + 1)),
                                     "width", GN_CARD_WIDTH,
                                     "height", GN_CARD_HEIGHT,
                                     NULL);

    gdk_pixbuf_unref (p);

    if (get_player (player)->type == GN_PLAYER_TYPE_HUMAN) {
        g_signal_connect (GTK_OBJECT (gci_new), "event",
                            GTK_SIGNAL_FUNC (gci_hand_cards_clicked_cb),
                            GINT_TO_POINTER (make_card_click_data (player, 0)));
    }

    g_ptr_array_add (get_player (player)->gci_hand_cards, gci_new);
}

void
gci_hand_cards_destroy (gint player)
{
    GnomeCanvasItem *gci;

    while (get_player (player)->gci_hand_cards->len > 0) {
        gci = GNOME_CANVAS_ITEM (g_ptr_array_remove_index (get_player (player)->gci_hand_cards, 0));
        gnome_canvas_item_hide (gci);
        gtk_object_destroy (GTK_OBJECT (gci));
    }

    g_ptr_array_free (get_player (player)->gci_hand_cards, TRUE);
    get_player (player)->gci_hand_cards = NULL;
}

void
gci_hand_cards_add (gint player, Card *card)
{
    GnomeCanvasItem *gci_new;
    GPtrArray *gci;
    GdkPixbuf *p;

    gci = get_player (player)->gci_hand_cards;

#ifndef DEBUG
    if (get_player (player)->type == GN_PLAYER_TYPE_COMPUTER) {
        p = card_get_picture (GN_COLOR_NONE, GN_CARD_BACK);
    } else {
#endif
        p = card_get_picture (card->color, card->type);
#ifndef DEBUG
    }
#endif

    if (get_player (player)->cards->len == 1) {
        gnome_canvas_item_set (GNOME_CANVAS_ITEM (g_ptr_array_index (gci, 0)), "pixbuf", p, NULL);
    } else {
        gci_new = gnome_canvas_item_new (gnome_canvas_root (gn_canvas),
                                         gnome_canvas_pixbuf_get_type (),
                                         "pixbuf", p,
                                         "x", (get_player (player)->cards->len * 20.0) - 10.0,
                                         "y", 30.0 + (105.0 * (player + 1.0)),
                                         "width", GN_CARD_WIDTH,
                                         "height", GN_CARD_HEIGHT,
                                         NULL);

        gdk_pixbuf_unref (p);

        g_ptr_array_add (gci, gci_new);

        if (get_player (player)->type == GN_PLAYER_TYPE_HUMAN) {
            g_signal_connect (GTK_OBJECT (gci_new), "event",
                                GTK_SIGNAL_FUNC (gci_hand_cards_clicked_cb),
                                GINT_TO_POINTER (make_card_click_data (player, get_player (player)->cards->len - 1)));
        }
    }
}

void
gci_hand_cards_remove (gint player, gint index)
{
    GdkPixbuf *p;
    GnomeCanvasItem *gci;
    gint c;

    if (get_player (player)->cards->len == 0) {
        gci = GNOME_CANVAS_ITEM (g_ptr_array_index (get_player (player)->gci_hand_cards, 0));
        p = card_get_picture (GN_COLOR_NONE, GN_CARD_EMPTY);
        gnome_canvas_item_set (gci, "pixbuf", p, NULL);
    } else {
        gci = GNOME_CANVAS_ITEM (g_ptr_array_remove_index (get_player (player)->gci_hand_cards, index));

        gtk_object_destroy (GTK_OBJECT (gci));

        for (c = index; c < get_player (player)->gci_hand_cards->len; c++) {
            gci = GNOME_CANVAS_ITEM (g_ptr_array_index (get_player (player)->gci_hand_cards, c));
            gnome_canvas_item_move (gci, -20.0, 0.0);
            if (get_player (player)->type == GN_PLAYER_TYPE_HUMAN) {
                g_signal_handlers_disconnect_matched (G_OBJECT (gci),
                                                      G_SIGNAL_MATCH_DATA,
                                                      0, 0, NULL, NULL,
                                                      GINT_TO_POINTER (make_card_click_data (player, c + 1)));
                g_signal_connect (GTK_OBJECT (gci), "event",
                                    GTK_SIGNAL_FUNC (gci_hand_cards_clicked_cb),
                                    GINT_TO_POINTER (make_card_click_data (player, c)));
            }
        }
    }
}

gint
gci_hand_cards_clicked_cb (GnomeCanvasItem *i, GdkEvent *event, gpointer data)
{
    gint p, c;

    if (!game_in_progress ()) return TRUE;

    if (event->type == GDK_BUTTON_PRESS) {
        p = (GPOINTER_TO_INT (data) & 240) >> 4;
        c = GPOINTER_TO_INT (data) & 15;
        if ((p == get_current_player ()) && (get_player (p)->cards->len != 0)) {
            if (valid_play (p, c))
                play_card (p, c);
            return TRUE;
        }
    }

    return FALSE;
}

gint
make_card_click_data (gint player, gint card)
{
    return (player << 4) + card;
}

void
gci_color_picker_hide ()
{
    gnome_canvas_item_hide (GNOME_CANVAS_ITEM (gn_pnl_color_picker));
}

void
gci_color_picker_show ()
{
    gnome_canvas_item_show (GNOME_CANVAS_ITEM (gn_pnl_color_picker));
}
