/*
 *  (Gnono) Card game for GNOME
 *
 *  Copyright (C) 2001-2002 Dennis J Houy <djhouy@paw.co.za>
 *            (C) 2006      Yaakov Selkowitz <yselkowitz@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *  $Id: players.c,v 1.7 2007/01/04 06:09:01 yselkowitz Exp $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/garray.h>
#include <glib/gtypes.h>
#include <glib/gi18n.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkwidget.h>
#include <libgnome/gnome-config.h>

#include "ai.h"
#include "cards.h"
#include "game.h"
#include "interface.h"
#include "players.h"

typedef enum {
    GN_CLOCKWISE      = 0,
    GN_ANTI_CLOCKWISE = -1
} GameDirections;

GtkWidget *playout;
static GPtrArray *players = NULL;
static gint gn_current_player = 1;
static GameDirections gn_direction = GN_CLOCKWISE;

void players_add_new (PlayerTypes type);

void
players_init () {
    gint p;
    gchar *key;
    gchar *playout_score;

    if (players == NULL) {
        playout = score_card_playout ();

        players = g_ptr_array_new ();

        players_add_new (GN_PLAYER_TYPE_COMPUTER);
        players_add_new (GN_PLAYER_TYPE_HUMAN);

        for (p = 0; p < get_num_players (); p++) {
            if (get_player (p)->type == GN_PLAYER_TYPE_COMPUTER) {
                key = g_strdup_printf ("Gnono/Names/AI=%s", _("Computer"));
            } else {
                key = g_strdup_printf ("Gnono/Names/%s=%s", g_get_user_name (), g_get_user_name ());
            }

            get_player (p)->name = gnome_config_private_get_string (key);

            g_free (key);

            get_player (p)->gci_hand_cards = NULL;
            get_player (p)->score_card = gn_score_card_new (p);
        }
    }

    playout_score = gnome_config_private_get_string("Gnono/GamePlay/PlayoutScore=100");
    gtk_label_set_text (GTK_LABEL (playout), playout_score);
    g_free (playout_score);

    for (p = 0; p < get_num_players (); p++) {
        player_cards_remove_all (p);
        gci_hand_cards_init (p);
        get_player (p)->score = 0;
        update_score_card (p);
    }

    gn_direction = GN_CLOCKWISE;
}

void
players_destroy ()
{
    gint p;
    ScoreCard *score_card;

    if (players != NULL) {

        gtk_widget_destroy (playout);

        for (p = 0; p < get_num_players (); p++) {
            score_card = get_player (p)->score_card;
            g_free (get_player (p)->name);
            get_player (p)->score_card = NULL;
            gtk_widget_destroy (GTK_WIDGET (score_card->total_points));
            g_free (score_card);

            player_cards_remove_all (p);
            g_ptr_array_free (get_player (p)->cards, TRUE);
            g_ptr_array_free (get_player (p)->gci_hand_cards, TRUE);
        }

        g_ptr_array_free (players, TRUE);
        players = NULL;
    }
}

void
players_add_new (PlayerTypes type) {
    Player *new_player = g_new (Player, 1);

    new_player->type = type;
    new_player->cards = g_ptr_array_new ();
    new_player->score = 0;

    g_ptr_array_add (players, new_player);
}

void
players_reverse_direction ()
{
    gn_direction = !gn_direction;
}

Player *
get_player (gint index)
{
    return (Player*) g_ptr_array_index (players, index);
}

void
player_cards_add (gint player, Card *card)
{
    g_ptr_array_add (get_player (player)->cards, card);
    gci_hand_cards_add (player, card);
}

Card *
player_cards_remove_index (gint player, gint index)
{
    Card *card;

    card = (Card*) g_ptr_array_remove_index (get_player (player)->cards, index);

    gci_hand_cards_remove (player, index);

    return card;
}

void
player_cards_remove_all (gint player)
{
    gint c, l;
    Card *card;

    l = get_player (player)->cards->len;

    for (c = 0; c < l; c++) {
        card = player_cards_remove_index (player, 0);
        g_free (card);
    }
}

gint
get_hand_score (gint player)
{
    Card *card;
    gint t = 0, c;

    for (c = 0; c < get_player (player)->cards->len; c++) {
        card = (Card*) g_ptr_array_index (get_player (player)->cards, c);

        switch (card->type) {
        case GN_CARD_1:
            t++;
            break;
        case GN_CARD_2:
            t += 2;
            break;
        case GN_CARD_3:
            t += 3;
            break;
        case GN_CARD_4:
            t += 4;
            break;
        case GN_CARD_5:
            t += 5;
            break;
        case GN_CARD_6:
            t += 6;
            break;
        case GN_CARD_7:
            t += 7;
            break;
        case GN_CARD_8:
            t += 8;
            break;
        case GN_CARD_9:
            t += 9;
            break;
        case GN_CARD_REVERSE:
            t += 20;
            break;
        case GN_CARD_SKIP:
            t += 20;
            break;
        case GN_CARD_TAKE_2:
            t += 20;
            break;
        case GN_CARD_WILD:
            t += 50;
            break;
        case GN_CARD_WILD_DRAW_4:
            t += 50;
            break;
        default:
            break;
        }
    }

    return t;
}

void
update_score_card (gint player)
{
    gchar *total;

    if (get_player (player)->score_card != NULL) {
        total = g_strdup_printf ("%d", get_player (player)->score);
        gtk_label_set_text (GTK_LABEL (get_player (player)->score_card->total_points), total);
        g_free (total);
    }
}

gint
get_current_player (void)
{
    return gn_current_player;
}

void
set_current_player (gint player)
{
    gn_current_player = player;
}

gint
next_player (gint player)
{
    if (gn_direction == GN_CLOCKWISE) {
        return player == (get_num_players () - 1) ? 0 : player + 1;
    } else {
        return player == 0 ? get_num_players () - 1 : player - 1;
    }
}
